using DynaWAVE
using NetalignMeasures
using NetalignUtils

function mainwave()
    tic()

    if length(ARGS) == 4
        net1 = ARGS[1]
        net2 = ARGS[2]
        simfile = ARGS[3]
        outputfile = ARGS[4]
    else
        println("Usage:
    julia dynawave.jl <network 1> <network 2> <node similarity file> <output alignment file>

    <network 1>, <network 2> --- Dynamic network files to align. For a given network file,
each line in the file describes one event with <start time> <end time> <node A> <node B>.
E.g.
<network 1>
3 5 a1 b1
6 7 a1 b1
0 2 a1 b1
0 4 b1 c1

<network 2>
0 2 a2 b2
3 5 a2 b2
6 7 a2 b2
0 4 b2 c2

<node similarity file> --- Contains node similarities between <network 1> and <network 2>.
Each line in the node similarity file contains the node similarity between a node in <network 1>
and a node in <network 2>.
E.g.
c1 a2 0.70
c1 b2 0.95
c1 f2 0.70
c1 e2 0.95
c1 c2 1.0

<output alignment file> --- Output alignment is written to this file. Each line in the alignment
file will contain a node mapping, from a node in <network 1> to a node in <network 2>.
E.g.
a1 a2
b1 b2
c1 c2
d1 d2
e1 f2

Example usage:
    julia dynawave.jl ../examples/ev1.txt ../examples/ev2.txt ../examples/evsim.txt output_alignment.txt
    julia dynawave.jl ../examples/yeastlc_original_tw_1_1.dy ../examples/yeastlc_rnd_0.10_1_tw_1_1.dy ../examples/yeastlc_yeastlc_nodesims.txt y_y_output_aln.txt
    ")
        return
    end

    println("Reading dynamic network files")
    t1 = readeventlist(net1)
    t2 = readeventlist(net2)
    S = readlistmat(simfile, t1.nodes, t2.nodes)

    println("Performing dynamic network alignment via DynaWAVE")
    if size(t1.G,1) <= size(t2.G,1)
        f, M = dynawave(t1.G, t2.G, S, details=true)
        flipg = false
    else
        f, M = dynawave(t2.G, t1.G, S', details=true)
        flipg = true
    end

    objval, objval_edge, objval_node = M.objval, M.nas.s.score, M.nas.t.score

    println("Writing output files")
    if outputfile == "-"
        fd = STDOUT
    else
        fd = open(outputfile, "w")
    end
    for i = 1:length(f)
        if !flipg
            println(fd, t1.nodes[i]," ", t2.nodes[f[i]])
        else
            println(fd, t1.nodes[f[i]]," ", t2.nodes[i])
        end
    end
    outputfile != "-" && close(fd)

    etime = toq()
    nc = !flipg ? nodecorrectness(f, t1.nodes, t2.nodes) : nodecorrectness(f, t2.nodes, t1.nodes)
    edgescore = !flipg ? score(DS3Measure(t1.G, t2.G),f) : score(DS3Measure(t2.G, t1.G),f)
    #nodescore = score(NodeSimMeasure(!flipg ? S : S'), f)

    # score 0.839640
    # node_score 0.000000
    # EC_score 0.919927
    # ICS_score 0.919927
    # S3_score 0.851727
    # DS3_score 0.839640
    # node_correctness 0.585657
    # elapsed_time 9538.000000
    println("score $(objval)")
    println("edge_score $(objval_edge)")
    println("node_score $(objval_node)")
    println("DS3_score $edgescore")
    println("node_correctness $nc")
    println("elapsed_time $etime")
end

mainwave()
